
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% demo / tutorial for matching code %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% For any question, email me: joan_glaunes@yahoo.fr

% In this example, we will match two sets composed of four pairs
% source/target of types surface, curve, unlabelled points and landmarks

% 1/ In a first variable "s", we input everything concerning the
% deformation part.

% s.x must be a 3*n array containing 3D coordinates of all source points.
% Here the surface is composed of two triangles (4 vertices), the source 
% curve is sampled with four points, the points set counts 2 points and there
% are 2 source landmarks, thus n=12
clear s
s.x = [0,1,1,0,1.5,1.6,1.8,2, 3,3.2,3.8, 4;
       0,0,1,1, 0 , 0 , 0 ,0, 0, 0 , 0 , 0 ;
       0,0,0,0, 0 , 0 , 0 ,0, 0, 0 , 0 , 0];
   
% s.sigmaV is the kernel size which defines the scale of deformation.
% In other words, the deformation kernel is a function of the euclidean
% distance between two points divided by sigmaV. One should fix a value
% corresponding to the range of coordinates of points. For example, 
% if data is inside a box of length D, sigmaV = 0.2*D or 0.5*D can be
% good choices.
s.sigmaV = .5;

% Other important parameters (see also match.m file)

s.rigidmatching = 0; % put = 1 to perform a rigid matching before elastic matching.

s.gammaR = 0;        % weight of deformation part in functional
% gammaR = 0 means we minimize only the matching term,
% but in this case the deformation will remain regular due to the space in
% which minimization is performed. To compute distances between
% objects, one should not set it to zero though.

s.numbminims = 1; % means we only perform one minimization. 
% Fixing for example numbminims=3 will tell the algorithm to do
% 3 consecutive minimizations, decreasing the size of the matching kernels
% sigmaW, sigmaI (see below) after each of them.

s.usefgt = 0; % setting 1 would mean we use Fast Gauss Transform for convolutions 
% (should be used when number of points involved is above 50 points approximately).




% 2/ In a second variable "target", we input everything concerning the
% targets and the matching methods used

clear target

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% First target of type surface
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

target{1}.method = 'surfcurr';

% Put indices of triangles of SOURCE surface in target{1}.vx
% These indices refer to corresponding columns of s.x
target{1}.vx = [1,1;
                2,3;
                3,4];
% Put coordinates of vertices of TARGET surface in target{1}.y
target{1}.y = [.5,1,0
                0,1,.5
               .5,1,.5];
% Put indices of triangles of TARGET surface in target{1}.vy            
target{1}.vy = [1;
                2;
                3];
% scale of surface matching kernel. Small values give more precision, but
% the algorithm may get stuck in local minima. So a good choice is a value
% corresponding to the gross distance between the source and target
% objects. However, if you use several minimizations (variable "numbminims",
% see above), then it is possible to fix a small value since the algorithm
% will perform the first steps at larger scales.
target{1}.sigmaW = .5;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Second target of type curve
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

target{2}.method = 'curvecurr';

% Put indices of segments of SOURCE curve on target{2}.vx
% These indices refer to corresponding columns of s.x
target{2}.vx = [5,6,7;
                6,7,8];
% Put coordinates of points of TARGET curve in target{2}.y
target{2}.y = [1.5, 2 ,2.5;
                0 , 0 ,0.5;
               0.5,0.5, 1 ];
% Put indices of segments of TARGET curve in target{2}.vy            
target{2}.vy = [1,2;
                2,3];
% scale of curve matching kernel
target{2}.sigmaW = .5;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Third target of type point set
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

target{3}.method = 'measures';

% Put indices of SOURCE points in target{3}.vx
% These indices refer to corresponding columns of s.x
target{3}.vx = [9,10];
% Put coordiantes of TARGET points in target{3}.y
target{3}.y = [3 , 3.2,3.4;
              0.5, 0.7,0.5;
               0.5,0.5,0.5];
% scale of point matching kernel
target{3}.sigmaI = 2;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fourth target of type landmarks
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

target{4}.method = 'landmarks';

% Put indices of SOURCE points in target{4}.vx
% These indices refer to crresponding columns of s.x
target{4}.vx = [11,12];
% Put coordinates of TARGET points in target{3}.y
target{4}.y = [ 4 , 4;
               0.5,0.5;
               0.5,0.8];



           
% Running the program:
s = match(s,target);

% The output structure "s" contains the variable "X" 
% (trajectories of template points) et "mom" (momentum vectors)
% which parameterize the optimal deformation. Also "distIdPhi" gives the
% distance to Identity in the space of diffeomorphisms.
disp(['deformation cost D = D(id,phi) = ',num2str(s.distIdPhi)])
disp(' ')

% display results
s.show = {'phi','y','xrig','x'};
s.showtraj = 1;    % display trajectories of source points (in blue)
s.showmomtraj = 0; % display momentum vectors (green arrows)
s.showpoints = 1;
s.showlegend = 0;
clf
affiche(s);
view(-20,20)
zoom(1.7)

% The program flow.m allows to compute the image of any 3D points
% under the optimal deformation map

disp('The''image of points')
V = rand(3,4)
disp('under the deformation map is')
phiV = flow(s,V)


