function s = measures(s)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%   Measure matching - to be used with match.m     %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% contact : joan_glaunes@yahoo.fr

method = 'measures';

% y (target points) is required
if ~isfield(s,'y')
    error('input structure must contain field <y> (target points)');
end

% y must be 3D data
if size(s.y,1)~=3
    error('field <y> must be 3-rows matrix');
end

y = s.y;
ny = size(y,2);

if isfield(s,'x')~=1
    if isfield(s,'nx')~=1
        if isfield(s,'wx')~=1
            if isfield(s,'vx')~=1
                error('at least one of fields x, nx, wx, vx should be given')
            else
                nx = length(s.vx);
            end
        else
            nx = length(s.wx);
        end
    else
        nx = s.nx;
    end
else
    nx = size(s.x,2);
end

if isfield(s,'wx')~=1
    wx = ones(1,nx)/nx;
else
    wx = s.wx;
end

if isfield(s,'wy')~=1
    wy = ones(1,ny)/ny;
else
    wy = s.wy;
end

if isfield(s,'vx')~=1
    vx = 1:nx;
else
    vx = s.vx;
end

numbminims = 1;
sigmaI = 'auto';            % matching kernel size
decsigmaI = 2.5;            % decrease rate of sigmaI after each minimization

usefgt = 0;                 % if 1 use IFGT algorithm
testfgt = 0;
order = 3;                  % IFGT order
K = min([10,floor((nx+ny)/2)]);  % IFGT number of clusters
e = 9;                      % IFGT cutoff ratio for targets

% update by input structure
loadstruct('s',who)

% auto scale
sclsz = mean(max(y')-min(y'));
if strcmp(sigmaI,'auto')
    sigmaI = .5*sclsz;
end
sigmaI2 = sigmaI^2;

premin = @funpremin;
matching = @funmatching;
gradmatching = @fungradmatching;

if usefgt
    IdScalProd = @fgtIdScalProd;
    fungrad = @fgtgradmatching;
else
    IdScalProd = @regIdScalProd;
    fungrad = @reggradmatching;
end

clear s
savestruct('s',who);

wz = [wx,-wy];

vxspec = 3 * vx;
vxspec = [vxspec-2;vxspec-1;vxspec];
vxspec= vxspec(:);

% multiscale settings
sigmaI = sigmaI*decsigmaI^numbminims;  %%% numbminims should be given by match.m
sigmaI2 = sigmaI^2;

% process this before each minimization
    function funpremin
        sigmaI = sigmaI/decsigmaI;
        sigmaI2 = sigmaI^2;
        if usefgt*testfgt
            essdir = regIdScalProd(y,wy,y,wy,sigmaI2);
            essfgt = fgtIdScalProd(y,wy,y,wy,sigmaI2);
            disp(['relative error fgt measures: ',num2str(abs((essdir-essfgt)/essdir))])
        end
    end

    function R = funmatching(phix)
        phix = reshape(phix,3,prod(size(phix))/3);        
        z = [phix(:,vx),y];
        R = IdScalProd(z,wz,z,wz,sigmaI2);
    end

    function R = regIdScalProd(a,wa,b,wb,sigmaI2)
        R = 0;
        na = size(a,2);
        nb = size(b,2);
        for k = 1:na
            Rk = 0;
            for j = 1:nb
                argin = -((a(1,k)-b(1,j))^2+(a(2,k)-b(2,j))^2+(a(3,k)-b(3,j))^2)/sigmaI2;
argout = exp(argin);  %% BUILT IN KERNEL kerI, do not remove this comment
                Rk = Rk + wb(j) * argout;
            end            
            R = R + wa(k) * Rk;
        end
    end

    function R = fgtIdScalProd(a,wa,b,wb,sigmaI2)
        resfgt = fgt(3,a,wa,b,sigmaI,order,K,e);
        R = sum(wb.*resfgt);
    end

    function eta = fungradmatching(phix)
        eta = fungrad(phix,sigmaI2);
    end

    function eta = reggradmatching(phix,sigmaI2)
        eta = zeros(size(phix));
        for e = 1:nx
            loce = 3*(vx(e)-1);
            eta1loc = 0;
            eta2loc = 0;
            eta3loc = 0;
            for f = 1:nx
                locf = 3*(vx(f)-1);
                argin = -(...
                    (phix(1+locf)-phix(1+loce))^2 + ...
                    (phix(2+locf)-phix(2+loce))^2 + ...
                    (phix(3+locf)-phix(3+loce))^2 )/sigmaI2;
argout = exp(argin);  %% BUILT IN KERNEL derI, do not remove this comment
                eta1loc = eta1loc + argout * wx(f) * (phix(1+locf)-phix(1+loce));
                eta2loc = eta2loc + argout * wx(f) * (phix(2+locf)-phix(2+loce));
                eta3loc = eta3loc + argout * wx(f) * (phix(3+locf)-phix(3+loce));
            end
            for f = 1:ny
                locf = 3*(f-1);
                argin = -(...
                    (y(1+locf)-phix(1+loce))^2 + ...
                    (y(2+locf)-phix(2+loce))^2 + ...
                    (y(3+locf)-phix(3+loce))^2 )/sigmaI2;
argout = exp(argin);  %% BUILT IN KERNEL derI, do not remove this comment
                eta1loc = eta1loc - argout * wy(f) * (y(1+locf)-phix(1+loce));
                eta2loc = eta2loc - argout * wy(f) * (y(2+locf)-phix(2+loce));
                eta3loc = eta3loc - argout * wy(f) * (y(3+locf)-phix(3+loce));
            end
            eta(1+loce) = 4 * wx(e) * eta1loc / sigmaI2;
            eta(2+loce) = 4 * wx(e) * eta2loc / sigmaI2;
            eta(3+loce) = 4 * wx(e) * eta3loc / sigmaI2;
        end
    end

    function eta = fgtgradmatching(phix,sigmaI2)
        sz = size(phix);
        phix = phix(vxspec);
        phix = reshape(phix,3,length(vx));
        z = [phix,y];        
        resfgt = fgt(3,z,[wz;specprod(z,wz,1)],phix,sigmaI,order,K,e);
        eta = zeros(3,prod(sz)/3);
        eta(:,vx) = (4/sigmaI2) * (specprod(resfgt(2:4,:),wx,1) - specprod(phix,wx.*resfgt(1,:),1));
        eta = reshape(eta,sz);
    end

end

function C = specprod(A,B,m)

n = size(A,2);
a = size(A,1);
b = size(B,1);

A = reshape(A,a/m,m,n);
B = reshape(B,m,b/m,n);
C = zeros(a/m,b/m,n);

for p = 1:n
    C(:,:,p) = A(:,:,p)*B(:,:,p);
end

C = reshape(C,a*b/m/m,n);

end


