function s = surfcurr(s)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Surface matching via currents - to be used with match.m %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

method = 'surfcurr';

% y (target points) is required
if ~isfield(s,'y')
    error('input structure must contain field <y> (target points)');
end

y = s.y;

% y must be 3D data
if size(y,1)~=3
    error('field <y> must be 3-rows matrix');
end

ny = size(y,2);

% vx and vy (template and target faces) are required components
if isfield(s,'vx')~=1 | isfield(s,'vy')~=1
    error('input structure must contain fields vx and vy (template and target faces)');
end

vx = s.vx;
vy = s.vy;
nfx = size(vx,2);         % number of template faces
nfy = size(vy,2);         % number of target faces

numbminims = 1;
sigmaW = 'auto';            % matching kernel size
decsigmaW = 2.5;            % decrease rate of sigmaW after each minimization

usefgt = 0;                 % if 1 use IFGT algorithm
testfgt = 0;
order = 5;                 % IFGT order
K = min([10,floor(ny/2)]);  % IFGT number of clusters
e = 9;                      % IFGT cutoff ratio for targets
wx = ones(1,nfx);
wy = ones(1,nfy);

% update by input structure
loadstruct('s',who)

% auto scale
sclsz = mean(max(y')-min(y'));
if strcmp(sigmaW,'auto')
    sigmaW = .5*sclsz;
end
sigmaW2 = sigmaW^2;

% computation of centers and normals for each face of target
[cy,Ny] = compcurr(y,vy,wy);

premin = @funpremin;
matching = @funmatching;
gradmatching = @fungradmatching;

clear s
savestruct('s',who);

% multiscale settings
sigmaW = sigmaW*decsigmaW^numbminims;
sigmaW2 = sigmaW^2;

if usefgt
    WdScalProd = @fgtWdScalProd;
    gradWdNorm = @fgtgradWdNorm;
else
    WdScalProd = @regWdScalProd;
    gradWdNorm = @reggradWdNorm;
end

% process this before each minimization
    function funpremin
        sigmaW = sigmaW/decsigmaW;
        sigmaW2 = sigmaW^2;
        if usefgt*testfgt
            essdir = regWdScalProd(cy,Ny,cy,Ny,sigmaW2);
            essfgt = fgtWdScalProd(cy,Ny,cy,Ny,sigmaW2);
            disp(['relative error fgt surfcurr: ',num2str(abs((essdir-essfgt)/essdir))])
        end
    end

    function R = funmatching(phix)
        [cphix,Nphix] = compcurr(phix,vx,wx); % computes centers and normals for each face
        R = WdScalProd([cphix,cy],[Nphix,-Ny],[cphix,cy],[Nphix,-Ny],sigmaW2);
    end


    function g = fungradmatching(phix)
        [cphix,Nphix] = compcurr(phix,vx,wx); % computation of centers and normals for each face
        g = gradWdNorm(phix,cphix,Nphix,sigmaW2);
    end


    function [ca,Na] = compcurr(a,va,wa)
        
        nfa = size(va,2);
        ca = zeros(3,nfa);
        Na = zeros(3,nfa);

        % computation of centers and normals for each face
        v = zeros(1,9);
        for f = 1:nfa
            locf = 3*(f-1);
            for k = 1:3
                for j = 1:3
                    v(k+3*(j-1)) = a(k+3*(va(j+locf)-1));
                end
            end
            % c = (v1+v2+v3)/3;
            ca(1+locf) = (v(1)+v(4)+v(7))/3;
            ca(2+locf) = (v(2)+v(5)+v(8))/3;
            ca(3+locf) = (v(3)+v(6)+v(9))/3;

            % N = [(v2-v1)a(v3-v1)]/2;
            Na(1+locf) = wa(f) * ((v(5)-v(2))*(v(9)-v(3))-(v(6)-v(3))*(v(8)-v(2)))/2;
            Na(2+locf) = wa(f) * ((v(6)-v(3))*(v(7)-v(1))-(v(4)-v(1))*(v(9)-v(3)))/2;
            Na(3+locf) = wa(f) * ((v(4)-v(1))*(v(8)-v(2))-(v(5)-v(2))*(v(7)-v(1)))/2;
        end

    end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Scalar product in W^* space %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function R = regWdScalProd(ca,Na,cb,Nb,sigmaW2)
        nfa = size(ca,2);
        nfb = size(cb,2);
        R = 0;
        for k = 1:nfa
            for j = 1:nfb
                argin = -((ca(1,k)-cb(1,j))^2+(ca(2,k)-cb(2,j))^2+(ca(3,k)-cb(3,j))^2)/sigmaW2;
argout = exp(argin);  %% BUILT IN KERNEL kerI, do not remove this comment
                R = R + argout * (Na(1,k)*Nb(1,j)+Na(2,k)*Nb(2,j)+Na(3,k)*Nb(3,j));
            end
        end
    end

    function R = fgtWdScalProd(ca,Na,cb,Nb,sigmaW2)
        R = sum(sum(Nb.*fgt(3,ca,Na,cb,sqrt(sigmaW2),order,K,e)));
    end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% gradient of W^* norm %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    function eta = reggradWdNorm(phi,cphi,Nphi,sigmaW2)

        eta = zeros(size(phi));
        etaloc = zeros(3);
        etaf_w= zeros(3,1);
        etaf_dw = zeros(3,1);
        v = zeros(1,9);

        for e = 1:nfx
            loce = 3*(e-1);
            etaf_w(1) = 0;
            etaf_w(2) = 0;
            etaf_w(3) = 0;
            etaf_dw(1) = 0;
            etaf_dw(2) = 0;
            etaf_dw(3) = 0;
            if wx(e)
            for f = 1:nfx
                locf = 3*(f-1);
                argin = -(...
                    (cphi(1+locf)-cphi(1+loce))^2 + ...
                    (cphi(2+locf)-cphi(2+loce))^2 + ...
                    (cphi(3+locf)-cphi(3+loce))^2 )/sigmaW2;
argout = exp(argin);  %% BUILT IN KERNEL kerI, do not remove this comment
                etaf_w(1) = etaf_w(1) + argout * Nphi(1+locf);
                etaf_w(2) = etaf_w(2) + argout * Nphi(2+locf);
                etaf_w(3) = etaf_w(3) + argout * Nphi(3+locf);
argout = exp(argin);  %% BUILT IN KERNEL derI, do not remove this comment
                dotN = Nphi(1+locf)*Nphi(1+loce) + Nphi(2+locf)*Nphi(2+loce) + Nphi(3+locf)*Nphi(3+loce);
                etaf_dw(1) = etaf_dw(1) + argout * dotN * (cphi(1+locf)-cphi(1+loce));
                etaf_dw(2) = etaf_dw(2) + argout * dotN * (cphi(2+locf)-cphi(2+loce));
                etaf_dw(3) = etaf_dw(3) + argout * dotN * (cphi(3+locf)-cphi(3+loce));
            end
            for f = 1:nfy
                locf = 3*(f-1);
                argin = -(...
                    (cy(1+locf)-cphi(1+loce))^2 + ...
                    (cy(2+locf)-cphi(2+loce))^2 + ...
                    (cy(3+locf)-cphi(3+loce))^2 )/sigmaW2;
argout = exp(argin);  %% BUILT IN KERNEL kerI, do not remove this comment
                etaf_w(1) = etaf_w(1) - argout * Ny(1+locf);
                etaf_w(2) = etaf_w(2) - argout * Ny(2+locf);
                etaf_w(3) = etaf_w(3) - argout * Ny(3+locf);
argout = exp(argin);  %% BUILT IN KERNEL derI, do not remove this comment
                dotN = Ny(1+locf)*Nphi(1+loce) + Ny(2+locf)*Nphi(2+loce) + Ny(3+locf)*Nphi(3+loce);
                etaf_dw(1) = etaf_dw(1) - argout * dotN * (cy(1+locf)-cphi(1+loce));
                etaf_dw(2) = etaf_dw(2) - argout * dotN * (cy(2+locf)-cphi(2+loce));
                etaf_dw(3) = etaf_dw(3) - argout * dotN * (cy(3+locf)-cphi(3+loce));
            end
            end
            etaf_w(1) = wx(e) * etaf_w(1);
            etaf_w(2) = wx(e) * etaf_w(2);
            etaf_w(3) = wx(e) * etaf_w(3);
            etaf_dw(1) = 4 * etaf_dw(1) / sigmaW2 / 3;
            etaf_dw(2) = 4 * etaf_dw(2) / sigmaW2 / 3;
            etaf_dw(3) = 4 * etaf_dw(3) / sigmaW2 / 3;
            for k = 1:3
                for j = 1:3
                    v(k+3*(j-1)) = phi(k+3*(vx(j+loce)-1));
                end
            end
            etaloc(1) = (v(5)-v(8))*etaf_w(3)-(v(6)-v(9))*etaf_w(2) + etaf_dw(1);
            etaloc(2) = (v(6)-v(9))*etaf_w(1)-(v(4)-v(7))*etaf_w(3) + etaf_dw(2);
            etaloc(3) = (v(4)-v(7))*etaf_w(2)-(v(5)-v(8))*etaf_w(1) + etaf_dw(3);
            etaloc(4) = (v(8)-v(2))*etaf_w(3)-(v(9)-v(3))*etaf_w(2) + etaf_dw(1);
            etaloc(5) = (v(9)-v(3))*etaf_w(1)-(v(7)-v(1))*etaf_w(3) + etaf_dw(2);
            etaloc(6) = (v(7)-v(1))*etaf_w(2)-(v(8)-v(2))*etaf_w(1) + etaf_dw(3);
            etaloc(7) = (v(2)-v(5))*etaf_w(3)-(v(3)-v(6))*etaf_w(2) + etaf_dw(1);
            etaloc(8) = (v(3)-v(6))*etaf_w(1)-(v(1)-v(4))*etaf_w(3) + etaf_dw(2);
            etaloc(9) = (v(1)-v(4))*etaf_w(2)-(v(2)-v(5))*etaf_w(1) + etaf_dw(3);
            loc1 = 3*(vx(1+loce)-1);
            loc2 = 3*(vx(2+loce)-1);
            loc3 = 3*(vx(3+loce)-1);
            eta(1+loc1) = eta(1+loc1) + etaloc(1);
            eta(2+loc1) = eta(2+loc1) + etaloc(2);
            eta(3+loc1) = eta(3+loc1) + etaloc(3);
            eta(1+loc2) = eta(1+loc2) + etaloc(4);
            eta(2+loc2) = eta(2+loc2) + etaloc(5);
            eta(3+loc2) = eta(3+loc2) + etaloc(6);
            eta(1+loc3) = eta(1+loc3) + etaloc(7);
            eta(2+loc3) = eta(2+loc3) + etaloc(8);
            eta(3+loc3) = eta(3+loc3) + etaloc(9);
        end
    end

    function eta = fgtgradWdNorm(phi,cphi,Nphi,sigmaW2)
        cz = [cphi,cy];
        Nz = [Nphi,-Ny];

        resfgt = fgt(3,cz,[Nz;specprod(cz,Nz,1)],cphi,sigmaW,order,K,e);

        eta_dw = (4/3/sigmaW2) * (specprod(resfgt(4:end,:),Nphi,3) - specprod(specprod(cphi,Nphi,1),resfgt(1:3,:),3));
        eta_w = resfgt(1:3,:);
        
        eta = zeros(prod(size(phi)),1);
        for m = 1:nfx
            locm = 3*m-2:3*m;
            vm1 = 3*vx(1,m); locvm1 = vm1-2:vm1;
            vm2 = 3*vx(2,m); locvm2 = vm2-2:vm2;
            vm3 = 3*vx(3,m); locvm3 = vm3-2:vm3;
            e1 = phi(locvm2)-phi(locvm3);
            e2 = phi(locvm3)-phi(locvm1);
            e3 = phi(locvm1)-phi(locvm2);
            eta(locvm1) = eta(locvm1)' + eta_dw(locm) + wx(m) * cross(e1,eta_w(locm));
            eta(locvm2) = eta(locvm2)' + eta_dw(locm) + wx(m) * cross(e2,eta_w(locm));
            eta(locvm3) = eta(locvm3)' + eta_dw(locm) + wx(m) * cross(e3,eta_w(locm));
        end
        eta = reshape(eta,size(phi));
    end


end

function C = specprod(A,B,m)

n = size(A,2);
a = size(A,1);
b = size(B,1);

A = reshape(A,a/m,m,n);
B = reshape(B,m,b/m,n);
C = zeros(a/m,b/m,n);

for p = 1:n
    C(:,:,p) = A(:,:,p)*B(:,:,p);
end

C = reshape(C,a*b/m/m,n);

end


